#pragma once

#include "base.h"
#include <numeric>

class consensus_testcase_adversary : public testcase_adversary_unison {
  public:
    consensus_testcase_adversary(int num_instances) :
      testcase_adversary_unison(num_instances), num_instances(num_instances),
      flipped(num_instances, false), partition(1), stat(2, 0) {
        for(int x = 0; x < num_instances; ++x) partition.back().push_back(x);
      }
    virtual ~consensus_testcase_adversary() {}

  protected:
    int num_instances;
    vector<char> flipped;
    vector<vector<int>> partition;
    vector<int> stat;

    vector<int> step(vector<int> S) {
      /*fprintf(stderr, "### partition is: ");
      for(auto& v : partition) {
        fprintf(stderr, "[");
        for(int x : v) fprintf(stderr, "%d ", x);
        fprintf(stderr, "] ");
      }*/

      vector<int> to_do;

      for(int i = 0; i < partition.size(); ++i)
        if(not singleStep(S, partition[i]))
          to_do.push_back(i);

      for(int i : to_do) fix(S, i);

      for(int x : S) ++stat[x];
      return S;
    }

    bool singleStep(vector<int> &S, vector<int> indices) {
      auto flips_to = [&](int b) {
        int f = 0;

        for(int i : indices) {
          if(flipped[i] and S[i] != b) f = num_instances + 1337;
          else f += (S[i] != b);
        }

        return make_pair(f, b);
      };

      auto [num_flips, b] = min(flips_to(0), flips_to(1));
      if(num_flips > num_instances) return false; // we're giving up

      for(int i : indices) {
        flipped[i] = (S[i] != b);
        S[i] = b;
      }

      return true;
    }

    void fix(vector<int> &S, int i) {
      /*
      fprintf(stderr, "to_do:");
      for(int i : to_do) fprintf(stderr, " %d", i);
      fprintf(stderr, "\n");*//*

      int num_ones = 0, num_zeroes = 0;
      for(int x : partition[i]) {
        if(S[x] == 0) ++num_ones;
        if(S[x] == 1) ++num_zeroes;
      }*/

      vector<vector<int>> split(2);

      auto want_to_set = [&](int index, int bit) {
        if(flipped[index]) {
          flipped[index] = false;
        }
        else {
          flipped[index] = (S[index] != bit);
          S[index] = bit;
        }

        split[S[index]].push_back(index);
      };

      int b = (stat[1] > stat[0] ? 1 : 0);
      for(int x : partition[i]) want_to_set(x, b);

      partition[i] = split[0];
      if(partition[i].empty())      partition[i] = split[1];
      else if(not split[1].empty()) partition.push_back(split[1]);
    }
};

class consensus_adversary : public adversary {
  public:
    consensus_adversary(int num_instances) : num_instances(num_instances) {}

    virtual void newRun() {
      assert(testcaseAdversary == 0);
      testcaseAdversary = new consensus_testcase_adversary(num_instances);
    }

  private:
    int num_instances;
};