#pragma once

#include "base.h"

class doublefusion_testcase_adversary : public testcase_adversary_unison {
  public:
    doublefusion_testcase_adversary(int num_instances, vector<int> &stat) :
      testcase_adversary_unison(num_instances), flipped(num_instances, false), stat(stat) {
        assert(num_instances == 4);
        assert(stat.size() == 2);
      }

  protected:
    vector<char> flipped;
    vector<int> &stat;

    bool singleStep(vector<int> &S, int o, int l) {
      assert(0 <= o and o < S.size());
      assert(0 <= l and l < S.size());
      assert(o != l);

      if(S[o] == S[l]) {
        flipped[o] = flipped[l] = false;
        return true;
      }

      if(flipped[o]) {
        flipped[o] = false; flipped[l] = true;
        S[l] = 1 - S[l];
        return true;
      }

      if(flipped[l]) {
        flipped[o] = true; flipped[l] = false;
        S[o] = 1 - S[o];
        return true;
      }

      return false;
    }

    vector<int> step(vector<int> S)
    {
      bool a = singleStep(S, 0, 1),
           b = singleStep(S, 2, 3);

      auto force = [this, &S](int i, int b) {
        flipped[i] = (S[i] != b);
        S[i] = b;
      };

      if(not a or not b) {
        int t = (stat[0] > stat[1] ? 0 : 1);
        if(a and S[0] != t) t = 1 - t;
        if(b and S[2] == t) t = 1 - t;
        if(not a) force(0, t), force(1, t);
        if(not b) force(2, 1 - t), force(3, 1 - t);
        assert(S[0] != S[2]);
        assert(S[0] == S[1] and S[2] == S[3]);
      }

      ++stat[S[0]];
      return S;
    }
};

class doublefusion_adversary : public adversary {
  public:
    doublefusion_adversary(int num_instances) :
      num_instances(num_instances), stat(2, 0), counter(-1), X(num_instances) {
        assert(num_instances == 4);
      }

    virtual pair<int, int> testcase(int T, int t, int MAX_N) {
      ++counter;

      if(counter % 4 == 0)
        X[0] = X[2] = rand(), X[1] = X[3] = rand();

      return {MAX_N, 1 + X[counter % 4] % MAX_N};
    }

    virtual void newRun() {
      assert(testcaseAdversary == 0);
      stat[0] = stat[1] = 0;
      testcaseAdversary = new doublefusion_testcase_adversary(num_instances, stat);
    }

  private:
    int num_instances;
    vector<int> stat;
    int counter;
    vector<int> X;
};