#pragma once

#include "base.h"

class fusion_testcase_adversary : public testcase_adversary_unison {
  public:
    fusion_testcase_adversary(int num_instances) :
      testcase_adversary_unison(num_instances), flipped(num_instances, false), stat(2, 0) {
        assert(num_instances == 2);
        assert(stat.size() == 2);
      }
    virtual ~fusion_testcase_adversary() {}

  protected:
    vector<char> flipped;
    vector<int> stat;

    vector<int> step(vector<int> S) {
      assert(S.size() == 2);
      assert(flipped.size() == 2);

      if(S[0] == S[1]) {
        flipped[0] = flipped[1] = false;
        ++stat[S[0]];
        return S;
      }

      if(flipped[0]) {
        flipped = {false, true};
        ++stat[S[0]];
        return {S[0], 1 - S[1]};
      }

      if(flipped[1]) {
        flipped = {true, false};
        ++stat[S[1]];
        return {1 - S[0], S[1]};
      }

      int b = (stat[0] > stat[1] ? 0 : 1);
      ++stat[b];
      flipped = {S[0] != b, S[1] != b};
      return {b, b};
    }
};

class fusion_adversary : public adversary {
  public:
    fusion_adversary(int num_instances, int period) : num_instances(num_instances), fixed(rand()), period(period) {
      assert(num_instances == 2);
    }

    virtual pair<int, int> testcase(int T, int t, int MAX_N) {
      static int counter = 0; ++counter;
      return {MAX_N, (counter % period ? rand() : fixed) % MAX_N + 1};
    }

    virtual void newRun() {
      assert(testcaseAdversary == 0);
      testcaseAdversary = new fusion_testcase_adversary(num_instances);
    }

  private:
    int num_instances;
    int fixed;
    int period;
};

class long_fusion_adversary : public adversary {
  public:
    long_fusion_adversary(int num_instances) : num_instances(num_instances), counter(-1), inputs(num_instances, -1) {
      refill();
    }

    virtual pair<int, int> testcase(int T, int t, int MAX_N) {
      ++counter;
      inputs[counter % num_instances] = 1 + extract_x() % MAX_N;
      //fprintf(stderr, "#%d", inputs[counter % num_instances]);
      return {MAX_N, inputs[counter % num_instances]};
    }

    virtual void newRun() {
      assert(testcaseAdversary == 0);
      testcaseAdversary = make_record_output_adv<fusion_testcase_adversary>(num_instances);
    }

    virtual void testcaseEnd() {
      if(testcaseAdversary) {
        auto adv = dynamic_cast<record_output_testcase_adversary *>(testcaseAdversary);
        int l = max(adv->getOutput(0).size(), adv->getOutput(1).size());
        for(int x : inputs) f[x] = max(f[x], l);
        //fprintf(stderr, "(%d) ", l);
      }

      adversary::testcaseEnd();
    }

  protected:
    static constexpr int NUM_ELEMS = 50;
    int counter;
    int num_instances;
    vector<int> X, Y;
    vector<int> inputs;
    unordered_map<int, int> f;

    void refill(int m = RAND_MAX) {
      while(X.size() < NUM_ELEMS) add(rand() % m, 0);
    }

    void add(int x, int w) {
      f[x] = w;
      X.push_back(x);
    }

    int extract_x() {
      if(X.empty()) next_gen();
      int x = X.back(); X.pop_back();
      Y.push_back(x);
      return x;
    }

    void next_gen() {
      sort(Y.begin(), Y.end(), [&](int i, int j) { return f[i] > f[j]; });
      X = Y;
      Y.clear();
      X.resize(X.size() / 2);
      refill();
    }
};