#pragma once

#include "base.h"

class fusion_backtrack_testcase_adversary : public testcase_adversary_unison {
  public:
    fusion_backtrack_testcase_adversary(int num_instances, vector<int> &choices) :
      testcase_adversary_unison(num_instances), flipped(num_instances, false), choices(choices), stat(2, 0), counter(-1) {
        assert(num_instances == 2);
      }

  protected:
    vector<char> flipped;
    vector<int> choices;
    vector<int> stat;
    int counter;

    vector<int> step(vector<int> S) {
      assert(S.size() == 2);
      assert(flipped.size() == 2);

      ++counter;

      if(S[0] == S[1]) {
        flipped[0] = flipped[1] = false;
        ++stat[S[0]];
        return S;
      }

      if(flipped[0]) {
        flipped = {false, true};
        ++stat[S[0]];
        return {S[0], 1 - S[1]};
      }

      if(flipped[1]) {
        flipped = {true, false};
        ++stat[S[1]];
        return {1 - S[0], S[1]};
      }

      int b = (stat[0] > stat[1] ? 0 : 1);
      if(counter < choices.size() and choices[counter] != -1) b = choices[counter];

      ++stat[b];
      flipped = {S[0] != b, S[1] != b};
      return {b, b};
    }
};

class fusion_backtrack_adversary : public adversary {
  public:
    fusion_backtrack_adversary(int num_instances) :
      num_instances(num_instances), input(2), counter(-1), best(0, vector<int>(333,-1)), temperature(142) {
        assert(num_instances == 2);
        for(int &x : input)  x = rand();
      }

    virtual pair<int, int> testcase(int T, int t, int MAX_N) {
      ++counter;
      return {MAX_N, 1 + input[counter % 2] % MAX_N};
    }

    virtual void testcaseEnd() {
      if(testcaseAdversary) {
        record_output_testcase_adversary *adv = dynamic_cast<record_output_testcase_adversary *>(testcaseAdversary);
        //assert(adv->getOutput(0) == adv->getOutput(1));
        //fprintf(stderr, "(%d)", (int) adv->getOutput(0).size());

        int l = max(adv->getOutput(0).size(), adv->getOutput(1).size());

        if(best.first <= l) {
          best.first = l;
          best.second = (l == adv->getOutput(0).size() ? adv->getOutput(0) : adv->getOutput(1));
          assert(best.first == best.second.size());
        }
      }

      adversary::testcaseEnd();
    }

    virtual void newRun() {
      assert(testcaseAdversary == 0);
      vector<int> choices = best.second;
      --temperature;

      if(not choices.empty()) {
        for(int i = 0; i < max(8, temperature / 2); ++i) {
          int idx = rand() % choices.size();
          choices[idx] = rand() % 2;
        }
      }

      testcaseAdversary = make_record_output_adv<fusion_backtrack_testcase_adversary>(num_instances, choices);

    }

  private:
    int num_instances;
    vector<int> choices;
    vector<int> input;
    int counter = 0;
    pair<int, vector<int>> best;
    int temperature;
};