#pragma once

#include <string>
#include <cassert>
#include <exception>
#include "random.h"
#include "force.h"
#include "pattern.h"
#include "combine.h"
#include "match.h"
#include "stat.h"
#include "fusion.h"
#include "fusion_force.h"
#include "doublefusion.h"
#include "majority.h"
#include "fusion_force.h"
#include "consensus.h"
#include "fusion_scatter.h"
#include "fractal.h"
#include "fusion_backtrack.h"
#include "majority_backtrack.h"

using namespace std;

adversary* create_adversary(FILE *fin, int num_instances) {
  auto readInt = [&fin]() {
    int x;
    assert(fscanf(fin, "%d", &x) == 1);
    return x;
  };
  auto readString = [&fin]() {
    char b[128];
    assert(fscanf(fin, "%s", b) == 1);
    return (string) b;
  };
  auto readPatterns = [&readInt, &readString]() {
    string type = readString();
    vector<vector<int>> p;
    int n = readInt();
    if (type == "explicit") {
      for (int i = 0; i < n; i++) {
        string s = readString();
        p.emplace_back();
        for (char c : s)
          p.back().push_back(c - '0');
      }
    } else {
      for (int b = readInt(); (int) p.size() < n; b++)
        for (int cp = 0; cp < (1 << b) && (int) p.size() < n; cp++) {
          bool m = true;
          for (int d = 1; d < b; d++)
            if (b % d == 0 && (cp & ((1 << (b - d)) - 1)) == (cp >> d))
              m = false;
          if (m) {
            p.emplace_back();
            for (int i = 0; i < b; i++)
              p.back().push_back((cp >> i) & 1);
          }
        }
    }
    return p;
  };
  string type = readString();
  if (type == "random")
    return new random_adversary();
  else if (type == "force")
    return new force_adversary(readPatterns());
  else if (type == "adaptive_force") {
    int X = readInt();
    auto patterns = readPatterns();
    return new adaptive_force_adversary(X, patterns);
  }
  else if (type == "dynamic_force") {
    int a = readInt(), b = readInt(), c = readInt();
    return new dynamic_force_adversary(a, b, c);
  }
  else if (type == "pattern")
    return new pattern_adversary(readPatterns());
  else if (type == "pattern_fixed") {
    int X = readInt();
    auto patterns = readPatterns();
    return new pattern_fixed_adversary(X, patterns);
  }
  else if (type == "combine_two")
    return new combine_two_adversary();
  else if (type == "combine_three")
    return new combine_three_adversary();
  else if (type == "match")
    return new match_adversary(readInt());
  else if (type == "stat")
    return new stat_adversary();
  else if (type == "fusion")
    return new fusion_adversary(num_instances, readInt());
  else if (type == "fusion_rotate")
    return new fusion_rotate_adversary(num_instances);
  else if (type == "doublefusion")
    return new doublefusion_adversary(num_instances);
  else if (type == "majority")
    return new majority_adversary(num_instances);
  else if (type == "fusion_rotate")
    return new fusion_rotate_adversary(num_instances);
  else if (type == "consensus")
    return new consensus_adversary(num_instances);
  else if (type == "fusion_scatter")
    return new fusion_scatter_adversary(num_instances);
  else if (type == "fractal")
    return new fractal_adversary(readInt());
  else if (type == "long_fusion")
    return new long_fusion_adversary(num_instances);
  else if (type == "fusion_backtrack")
    return new fusion_backtrack_adversary(num_instances);
  else if (type == "majority_backtrack")
    return new majority_backtrack_adversary(num_instances);
  else
    throw logic_error(type);
}