#pragma once

#include "base.h"

class majority_testcase_adversary : public testcase_adversary_unison {
  public:
    majority_testcase_adversary(int num_instances, vector<int> &stat) :
      testcase_adversary_unison(num_instances), flipped(num_instances, false), special(-1), stat(stat) {
        assert(num_instances == 3);
      }

  protected:
    vector<char> flipped;
    int special;
    vector<int> &stat;

    vector<int> step(vector<int> S) {

      assert(num_instances == 3);
      assert(S.size() == num_instances);
      assert(flipped.size() == num_instances);
      assert(phases.size() == num_instances);

      for(int i = 0; i < num_instances; ++i)
        if(phases[i] == TIRED and special == -1) special = i;

      if(special != -1) return fusionStep(S);

      int minority = -1;
      for(int i = 0; i < num_instances; ++i)
        if(S[i] != S[a(i)] and S[i] != S[b(i)])
          minority = i;

      if(minority == -1) {
        flipped = {false, false, false};
        for(int b : S) ++stat[b];
        return S;
      }

      auto flip = [&S, this](int i) {
        S[i] = 1 - S[i];
        flipped[i] = true;
      };

      if(flipped[minority]) {
        if(flipped[a(minority)] or flipped[b(minority)]) {
          assert(special == -1);
          special = minority;
          return fusionStep(S);
        }

        flipped = {false, false, false};
        flip(a(minority));
        flip(b(minority));
      }

      else {
        flipped = {false, false, false};
        flip(minority);
      }

      for(int b : S) ++stat[b];
      return S;
    }

    static int a(int x) { return (x + 1) % 3; }
    static int b(int x) { return (x + 2) % 3; }

    vector<int> fusionStep(vector<int> S) {
      assert(special != -1);
      int i = a(special), j = b(special), k = special;

      if(S[i] == S[j]) {
        flipped[i] = flipped[j] = false;
      }

      else if(flipped[i]) {
        int b = S[j];
        if(not flipped[j]) b = S[i];
        flipped[i] = false;
        flipped[j] = (S[i] != S[j]);
        S[j] = b;

      }

      else {
        flipped[j] = false;
        int b = S[i];
        if(not flipped[i]) b = S[j];
        flipped[i] = (S[i] != S[j]);
        S[i] = b;
      }

      //assert(S[i] == S[j]);

      if(flipped[k]) {
        flipped[k] = false;
      }

      else {
        int target = (stat[0] > stat[1] ? 0 : 1);
        flipped[k] = (target != S[k]);
        S[k] = target;
      }

      for(int b : S) ++stat[b];
      return S;
    }
};

class majority_adversary : public adversary {
  public:
    majority_adversary(int num_instances) :
      num_instances(num_instances), stat(2, 0) { assert(num_instances == 3); }


    virtual void newRun() {
      assert(testcaseAdversary == 0);
      testcaseAdversary = new majority_testcase_adversary(num_instances, stat);
    }

  private:
    int num_instances;
    vector<int> stat;
};