#pragma once

#include "base.h"

class majority_backtrack_testcase_adversary : public testcase_adversary_unison {
  public:
    majority_backtrack_testcase_adversary(int num_instances, vector<int> choices) :
      testcase_adversary_unison(num_instances), flipped(num_instances, false), special(-1), stat(2), counter(-1) {
        assert(num_instances == 3);
      }

  protected:
    vector<char> flipped;
    int special;
    vector<int> stat;
    vector<int> choices;
    int counter;

    vector<int> step(vector<int> S) {
      assert(num_instances == 3);
      assert(S.size() == num_instances);
      assert(flipped.size() == num_instances);
      assert(phases.size() == num_instances);

      ++counter;

      for(int i = 0; i < num_instances; ++i)
        if(phases[i] == TIRED and special == -1) special = i;

      if(special != -1) return fusionStep(S);

      int minority = -1;
      for(int i = 0; i < num_instances; ++i)
        if(S[i] != S[a(i)] and S[i] != S[b(i)])
          minority = i;

      if(minority == -1) {
        flipped = {false, false, false};
        for(int b : S) ++stat[b];
        return S;
      }

      auto flip = [&S, this](int i) {
        S[i] = 1 - S[i];
        flipped[i] = true;
      };

      if(flipped[minority]) {
        if(flipped[a(minority)] or flipped[b(minority)]) {
          assert(special == -1);
          special = minority;
          return fusionStep(S);
        }

        flipped = {false, false, false};
        flip(a(minority));
        flip(b(minority));
      }

      else {
        flipped = {false, false, false};
        flip(minority);
      }

      for(int b : S) ++stat[b];
      return S;
    }

    static int a(int x) { return (x + 1) % 3; }
    static int b(int x) { return (x + 2) % 3; }

    vector<int> fusionStep(vector<int> S) {
      assert(special != -1);
      int i = a(special), j = b(special), k = special;

      if(S[i] == S[j]) {
        flipped[i] = flipped[j] = false;
      }

      else if(flipped[i]) {
        assert(not flipped[j]);
        flipped[i] = false;
        flipped[j] = (S[i] != S[j]);
        S[j] = S[i];

      }

      else {
        flipped[j] = false;
        flipped[i] = (S[i] != S[j]);
        S[i] = S[j];
      }

      assert(S[i] == S[j]);

      if(flipped[k]) {
        flipped[k] = false;
      }

      else {
        int target = S[i];
        flipped[k] = (target != S[k]);
        S[k] = target;
      }

      for(int b : S) ++stat[b];
      return S;
    }
};

class majority_backtrack_adversary : public adversary {
  public:
    majority_backtrack_adversary(int num_instances) :
      num_instances(num_instances), counter(-1), choices(333), inputs(num_instances) {
        assert(num_instances == 3);
        for(int &x : inputs) x = rand();
        for(int &x : choices) x = rand() % 4;
      }

    virtual pair<int, int> testcase(int MAX_N) {
      ++counter;
      return {MAX_N, 1 + inputs[counter % 3] % MAX_N};
    }

    virtual void newRun() {
      assert(testcaseAdversary == 0);
      testcaseAdversary = new majority_testcase_adversary(num_instances, choices);
    }

  private:
    int num_instances;
    int counter;
    vector<int> choices;
    vector<int> inputs;
};